import axios from 'axios'
import database from './database.js'
import { execSync } from 'child_process'
import express from 'express'
import formidable from 'express-formidable'
import fs from 'fs'
import pug from 'pug'
import stylus from 'stylus'
import util from 'util'

let cms = {}

let ports = []
if (process.argv.indexOf('--ports') != -1) ports = process.argv[process.argv.indexOf('--ports') + 1].split(',')
let cmsPort = process.argv[process.argv.indexOf('--cms-port') + 1]
let cmsUrl = `http://localhost:${cmsPort}`
let addonId = process.argv[process.argv.indexOf('--addon-id') + 1]
let developmentApiKey = ''
let developmentCmsUrl = ''

let eventCallbacks = {}

// get kebab-case and camelCase names
cms.getHtmlName = s => s.replace(/[ _\.\/]/gi, '-').replace(/[A-Z]+/g, s => `-${s.toLowerCase()}`).replace(/[^a-z0-9\-]/g, '').replace(/-+/g, '-').replace(/^-/, '').replace(/-$/, '')
cms.getJsName = s => cms.getHtmlName(s).replace(/-./g, s => s.slice(1).toUpperCase())

// build site
cms.build = async (options={}) => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/build`, {env: options.env == 'production' ? 'production' : 'development'})
}

// set up database
database.setUrl(`${cmsUrl}/api/addon/${addonId}/database`)
cms.database = database

// create a page
cms.createPage = async page => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/page`, page)
  return response.data.pageDraft
}

// clear all data items in datastore
cms.clearDatastore = async options => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/datastore/${options.datastoreId}/clear`)
}

// create media
cms.createMedia = async media => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/media`, media)
  return response.data.media
}

// create schedule
cms.createSchedule = async schedule => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/schedule`, schedule)
  return response.data.schedule
}

// create server
cms.createServer = (...args) => {
  let app = express(...args)
  app.use(formidable())

  // create appProxy to allow catching async errors in controllers
  let appProxy = {
    use: (...args) => app.use(...args),
  }
  for (let method of ['all', 'get', 'put', 'post', 'delete']) {
    appProxy[method] = (...args) => {
      let originalHandler = args[args.length-1]
      args[args.length-1] = async (req, res, next) => {
        try {
          await originalHandler(req, res, next)
        } catch (err) {
          next(err)
        }
      }
      app[method](...args)
    }
  }
  appProxy.listen = (...args) => app.listen(...args)
  appProxy.listenToEvents = eventPath => {
    app.post(eventPath, async (req, res) => {
      let hasSent = false
      let shouldWait = false
      let event = req.fields.event
      event.status = status => req.status(status)
      event.send = response => {
        if (hasSent) return
        hasSent = true
        res.send(response)
      }
      event.wait = () => shouldWait = true
      let eventLog = ''
      if (event.type == 'scheduleStarted') {
        event.send()
        event.log = (...args) => args.forEach(arg => eventLog += `${eventLog.length ? '\n' : ''}${(arg && typeof arg == 'object') ? util.inspect(arg) : arg}`)
        event.complete = async (result={}) => await axios.post(`${cmsUrl}/api/addon/${addonId}/schedule/${event.scheduleId}/complete`, {runId: event.runId, status: result.status, log: result.log})
        event.success = async (...args) => {
          if (args.length) event.log(...args)
          await event.complete({status: 'success', log: eventLog})
        }
        event.error = async (...args) => {
          if (args.length) event.log(...args)
          await event.complete({status: 'error', log: eventLog})
        }
      }
      if (!eventCallbacks[event.type]) eventCallbacks[event.type] = []
      if (event.type == 'scheduleStarted') {
        if (!eventCallbacks[`${event.type}:${event.eventName}`]) eventCallbacks[`${event.type}:${event.eventName}`] = []
        for (let callback of eventCallbacks[`${event.type}:${event.eventName}`]) {
          try {
            await callback(event)
          } catch (err) {
            console.log(err)
            return res.status(500).send()
          }
        }
      }
      for (let callback of eventCallbacks[event.type]) {
        try {
          await callback(event)
        } catch (err) {
          console.log(err)
          return res.status(500).send()
        }
      }
      if (!hasSent && !shouldWait) res.send()
    })
  }
  appProxy.on = (eventName, callback) => {
    if (!eventCallbacks[eventName]) eventCallbacks[eventName] = []
    eventCallbacks[eventName].push(callback)
    return callback
  }
  appProxy.off = (eventName, callback) => {
    if (!eventCallbacks[eventName]) eventCallbacks[eventName] = []
    while (eventCallbacks[eventName].includes(callback)) eventCallbacks[eventName].splice(eventCallbacks.indexOf(eventName), 1)
    return callback
  }

  return appProxy
}

// delete components
cms.deleteComponents = async componentIds => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/component/delete`, {componentIds})
}

// delete data items that match
cms.deleteDataItems = async options => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/dataitem/${options.datastoreId}/delete`, {match: options.match})
}

// delete datastores
cms.deleteDatastores = async datastoreIds => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/datastore/delete`, {datastoreIds})
}

// delete repoPages
cms.deleteRepoPages = async repoPageIds => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/repopage/delete`, {repoPageIds})
}

// delete repoLayouts
cms.deleteRepoLayouts = async repoLayoutIds => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/repolayout/delete`, {repoLayoutIds})
}

// disable dataitems
cms.disableDataItems = async options => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/dataitem/${options.datastoreId}/disable`, {dataItemIds: options.dataItemIds, notDataItemIds: options.notDataItemIds, match: options.match, publishComment: options.publishComment})
}

// disable dataitems
cms.enableDataItems = async options => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/dataitem/${options.datastoreId}/enable`, {dataItemIds: options.dataItemIds, notDataItemIds: options.notDataItemIds, match: options.match, publishComment: options.publishComment})
}

// serve static content from directory
cms.static = (...args) => express.static(...args)

// get addon id
cms.getAddonId = () => addonId

// get name of addon
cms.getAddonName = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.state.name
}

// get path to assets
cms.getAssetPath = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.cms.assetPath
}

// get assigned cms ports
cms.getCmsPort = () => cmsPort
cms.getPort = () => ports[0]
cms.getPorts = () => ports

// get cms state
cms.getCmsState = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.cms 
}

// get timezone
cms.getTimezone = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.cms.timezone
}

// get isTrailingSlashEnabled
cms.isTrailingSlashEnabled = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.cms.isTrailingSlashEnabled
}

// get site name
cms.getSiteName = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.cms.name
}

// get addon's components
cms.getComponents = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/component`)
  return response.data.components
}

// get data item by id
cms.getDataItem = async options => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/dataitem/${options.datastoreId}/${options.dataItemId}`)
  return response.data.dataItem
}

// get data items
cms.getDataItems = async options => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/dataitem/${options.datastoreId}/list`, {dataItemIds: options.dataItemIds, match: options.match})
  return response.data.dataItems
}

// get addon's datastores
cms.getDatastores = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/datastore`)
  return response.data.datastores
}

// get help
cms.getHelp = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.state.help
}

// get addon's media
cms.getMedia = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/media`)
  return response.data.media
}

// get published pages
cms.getPages = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/page`)
  return response.data.pages
}

// get page drafts
cms.getPageDrafts = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/page`)
  return response.data.pageDrafts
}

// get addon's repoPages
cms.getRepoPages = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/repopage`)
  return response.data.repoPages
}

// get addon's repoLayouts
cms.getRepoLayouts = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/repolayout`)
  return response.data.repoLayouts
}

// get addon's schedules
cms.getSchedules = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/schedule`)
  return response.data.schedules
}

// get sidebar items
cms.getSidebarItems = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.state.sidebarItems
}

// get addon state
cms.getState = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.state
}

// get storage path
cms.getStoragePath = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.state.storagePath
}

cms.getCodePath = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.state.codePath
}

// get url to talk to cms
cms.getUrl = () => cmsUrl

// get cms uuid
cms.getUuid = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.cms.uuid
}

// get whether development mode is enabled in the cms
cms.isDevelopmentModeEnabled = async () => {
  let response = await axios.get(`${cmsUrl}/api/addon/${addonId}/state`)
  return response.data.cms.isDevelopmentModeEnabled
}

// set help in the interface
cms.setHelp = async help => {
  for (let helpPage of help) {
    if (helpPage.bodyFile) {
      helpPage.body = fs.readFileSync(helpPage.bodyFile, 'utf-8')
      delete helpPage.bodyFile
    }
  }
  await axios.put(`${cmsUrl}/api/addon/${addonId}/state/help`, {help})
}

// set sidebar items in the interface
cms.setSidebarItems = async sidebarItems => {
  await axios.put(`${cmsUrl}/api/addon/${addonId}/state/sidebaritems`, {sidebarItems})
}

// update or create component
cms.upsertComponent = async (match, component, publishComment) => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/component/upsert`, {match, component, publishComment})
  return response.data.component
}

// update or create a single dataitem
cms.upsertDataItem = async (options) => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/dataitem/${options.datastoreId}/upsert`, {match: options.match, dataItem: options.dataItem, raw: options.raw, publishComment: options.publishComment})
  return response.data.dataItem
}

// update or create dataitems
cms.upsertDataItems = async (options) => {
  let dataItems = options.dataItems.slice()
  let createdDataItems = []
  while (dataItems.length) {
    let dataItem = dataItems.shift()
    let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/dataitem/${options.datastoreId}/upsert`, {match: options.match, dataItem, raw: options.raw, publishComment: options.publishComment})
    createdDataItems.push(response.data.dataItem)
  }
  return createdDataItems
}

// update or create a datastore
cms.upsertDatastore = async (match, datastore, publishComment) => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/datastore/upsert`, {match, datastore, publishComment})
  return response.data.datastore
}

// update or create a repo layout
cms.upsertRepoLayout = async (match, repoLayout) => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/repolayout/upsert`, {match, repoLayout})
  return response.data.repoLayout
}

// update or create a repo page
cms.upsertRepoPage = async (match, repoPage) => {
  let response = await axios.post(`${cmsUrl}/api/addon/${addonId}/repopage/upsert`, {match, repoPage})
  return response.data.repoPage
}

// build directory
cms.buildDirectory = async (from, to, options) => {
  if (!to) throw 'must specify output directory'
  if (options.clear) fs.rmSync(to, {recursive: true, force: true})
  fs.mkdirSync(to, {recursive: true, force: true})
  let parseDirectory = async dir => {
    let outputPath = `${to}/${dir.slice(from.length)}`
    fs.mkdirSync(outputPath, {recursive: true})
    let items = fs.readdirSync(dir)
    for (let item of items) {
      let inputPath = `${dir}/${item}`
      if (fs.lstatSync(inputPath).isDirectory()) {
        await parseDirectory(inputPath)
        continue
      }
      let itemName = item.split('.').slice(0, -1).join('.')
      if (item.endsWith('.pug')) {
        fs.writeFileSync(`${outputPath}/${itemName}.html`, pug.compile(fs.readFileSync(inputPath, 'utf-8'))())
      } else if (item.endsWith('.styl')) {
        fs.writeFileSync(`${outputPath}/${itemName}.css`, stylus.render(fs.readFileSync(inputPath, 'utf-8')))
      } else {
        fs.cpSync(inputPath, `${outputPath}/${item}`)
      }
    }
  }
  await parseDirectory(from)
}

// install an addon package
cms.installAddonPackage = async options => {
  try {
    await axios.post(`${developmentCmsUrl}/api/development/addon/install`, {
        file: fs.createReadStream(options.path)
      }, {
      headers: {
        'Authorization': options.developmentApiKey || developmentApiKey,
        'Content-Type': 'multipart/form-data'
      },
    })
  } catch (err) {
    console.log('Development API Key Invalid')
    if (err && err.response && err.response.data && err.response.data == 'developmentApiKeyInvalid') throw 'developmentApiKeyInvalid'
    throw err
  }
}

// load a component from a zip
cms.loadComponentPackage = path => {
  let data
  if (fs.lstatSync(path).isDirectory()) data = fs.readFileSync(`${path}/component.json`)
  else data = execSync(`unzip -p "${path}" 'component.json'`)
  let component = JSON.parse(data)
  return {
    component,
  }
}

// load a datastore from a zip
cms.loadDatastorePackage = path => {
  let data
  if (fs.lstatSync(path).isDirectory()) data = fs.readFileSync(`${path}/datastore.json`)
  else data = execSync(`unzip -p "${path}" 'datastore.json'`)
  let datastore = JSON.parse(data)
  return {
    datastore,
  }
}

// load a page from a zip
cms.loadPagePackage = path => {
  let data
  if (fs.lstatSync(path).isDirectory()) data = fs.readFileSync(`${path}/page.json`)
  else data = execSync(`unzip -p "${path}" 'page.json'`)
  let page = JSON.parse(data)
  return {
    page
  }
}

// load a repoPage from a zip
cms.loadRepoPagePackage = path => {
  let data
  if (fs.lstatSync(path).isDirectory()) data = fs.readFileSync(`${path}/page.json`)
  else data = execSync(`unzip -p "${path}" 'page.json'`)
  let page = JSON.parse(data)
  return {
    page,
  }
}

// load a repoLayout from a zip
cms.loadRepoLayoutPackage = path => {
  let data
  if (fs.lstatSync(path).isDirectory()) data = fs.readFileSync(`${path}/layout.json`)
  else data = execSync(`unzip -p "${path}" 'layout.json'`)
  let layout = JSON.parse(data)
  return {
    layout,
  }
}

// move a media item
cms.moveMedia = async (mediaIds, path) => {
  if (!Array.isArray(mediaIds)) mediaIds = [mediaIds]
  await axios.post(`${cmsUrl}/api/addon/${addonId}/media/move`, {mediaIds, path})
} 

// package the addon for installation
cms.packageAddon = async (options={}) => {
  let outputPath = options.outputPath
  let inputPath = options.inputPath || process.cwd()
  let install = options.install
  if (!outputPath) {
    let manifest = JSON.parse(fs.readFileSync('addon.json'))
    outputPath = `../${manifest.name}-${manifest.version.replace(/\./g, '-')}.addon.zip`
  }
  if (!outputPath.startsWith('/')) outputPath = `${process.cwd()}/${outputPath}`
  if (!outputPath.endsWith('.addon.zip')) outputPath = `${outputPath}/.addon.zip`
  fs.rmSync(outputPath, {recursive: true, force: true})
  execSync(`cd ${inputPath} && zip -r ${outputPath} * -x node_modules/\\*`)
  if (options.install) await cms.installAddonPackage({path: outputPath, developmentApiKey: options.developmentApiKey, developmentCmsUrl: options.developmentCmsUrl})
}

// show notification in interface for all users
cms.queueNotification = async options => {
  await axios.post(`${cmsUrl}/api/addon/${addonId}/notification`, {
    body: options.body,
  })
}

// set development mode options
cms.setDevelopmentOptions = options => {
  if ('developmentApiKey' in options) developmentApiKey = options.developmentApiKey
  if ('developmentCmsUrl' in options) developmentCmsUrl = options.developmentCmsUrl
}

export default cms
